/* Copyright (C) 2004 - 2005  db4objects Inc.  http://www.db4o.com

This file is part of the db4o open source object database.

db4o is free software; you can redistribute it and/or modify it under
the terms of version 2 of the GNU General Public License as published
by the Free Software Foundation and as clarified by db4objects' GPL 
interpretation policy, available at
http://www.db4o.com/about/company/legalpolicies/gplinterpretation/
Alternatively you can write to db4objects, Inc., 1900 S Norfolk Street,
Suite 350, San Mateo, CA 94403, USA.

db4o is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. */
namespace com.db4o
{
	internal class PendingClassInits
	{
		private readonly com.db4o.YapClassCollection _classColl;

		private com.db4o.foundation.Collection4 _pending = new com.db4o.foundation.Collection4
			();

		private com.db4o.foundation.Queue4 _members = new com.db4o.foundation.Queue4();

		private com.db4o.foundation.Queue4 _statics = new com.db4o.foundation.Queue4();

		private com.db4o.foundation.Queue4 _writes = new com.db4o.foundation.Queue4();

		private com.db4o.foundation.Queue4 _inits = new com.db4o.foundation.Queue4();

		private bool _running = false;

		internal PendingClassInits(com.db4o.YapClassCollection classColl)
		{
			_classColl = classColl;
		}

		internal virtual void process(com.db4o.YapClass newYapClass)
		{
			if (_pending.contains(newYapClass))
			{
				return;
			}
			com.db4o.YapClass ancestor = newYapClass.getAncestor();
			if (ancestor != null)
			{
				process(ancestor);
			}
			_pending.add(newYapClass);
			_members.add(newYapClass);
			if (_running)
			{
				return;
			}
			_running = true;
			checkInits();
			_pending = new com.db4o.foundation.Collection4();
			_running = false;
		}

		private void checkMembers()
		{
			while (_members.hasNext())
			{
				com.db4o.YapClass yc = (com.db4o.YapClass)_members.next();
				yc.addMembers(_classColl.i_stream);
				_statics.add(yc);
			}
		}

		private void checkStatics()
		{
			checkMembers();
			while (_statics.hasNext())
			{
				com.db4o.YapClass yc = (com.db4o.YapClass)_statics.next();
				yc.storeStaticFieldValues(_classColl.i_systemTrans, true);
				_writes.add(yc);
				checkMembers();
			}
		}

		private void checkWrites()
		{
			checkStatics();
			while (_writes.hasNext())
			{
				com.db4o.YapClass yc = (com.db4o.YapClass)_writes.next();
				yc.setStateDirty();
				yc.write(_classColl.i_systemTrans);
				_inits.add(yc);
				checkStatics();
			}
		}

		private void checkInits()
		{
			checkWrites();
			while (_inits.hasNext())
			{
				com.db4o.YapClass yc = (com.db4o.YapClass)_inits.next();
				yc.initConfigOnUp(_classColl.i_systemTrans);
				checkWrites();
			}
		}
	}
}
