/* Copyright (C) 2004 - 2005  db4objects Inc.  http://www.db4o.com

This file is part of the db4o open source object database.

db4o is free software; you can redistribute it and/or modify it under
the terms of version 2 of the GNU General Public License as published
by the Free Software Foundation and as clarified by db4objects' GPL 
interpretation policy, available at
http://www.db4o.com/about/company/legalpolicies/gplinterpretation/
Alternatively you can write to db4objects, Inc., 1900 S Norfolk Street,
Suite 350, San Mateo, CA 94403, USA.

db4o is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. */
namespace com.db4o
{
	internal class YLong : com.db4o.YapJavaClass
	{
		private static readonly long i_primitive = System.Convert.ToInt64(0);

		public YLong(com.db4o.YapStream stream) : base(stream)
		{
		}

		public override object coerce(com.db4o.reflect.ReflectClass claxx, object obj)
		{
			return com.db4o.foundation.Coercion4.toLong(obj);
		}

		public override object defaultValue()
		{
			return i_primitive;
		}

		public override int getID()
		{
			return 2;
		}

		protected override j4o.lang.Class primitiveJavaClass()
		{
			return j4o.lang.Class.getClassForType(typeof(long));
		}

		public override int linkLength()
		{
			return com.db4o.YapConst.YAPLONG_LENGTH;
		}

		internal override object primitiveNull()
		{
			return i_primitive;
		}

		internal override object read1(com.db4o.YapReader a_bytes)
		{
			long ret = readLong(a_bytes);
			return ret;
		}

		internal static long readLong(com.db4o.YapReader a_bytes)
		{
			long l_return = 0;
			for (int i = 0; i < com.db4o.YapConst.LONG_BYTES; i++)
			{
				l_return = (l_return << 8) + (a_bytes._buffer[a_bytes._offset++] & unchecked((int
					)(0xff)));
			}
			return l_return;
		}

		public override void write(object a_object, com.db4o.YapWriter a_bytes)
		{
			if (!com.db4o.Deploy.csharp && a_object == null)
			{
				writeLong(long.MaxValue, a_bytes);
			}
			else
			{
				writeLong(((long)a_object), a_bytes);
			}
		}

		internal static void writeLong(long a_long, com.db4o.YapWriter a_bytes)
		{
			for (int i = 0; i < com.db4o.YapConst.LONG_BYTES; i++)
			{
				a_bytes._buffer[a_bytes._offset++] = (byte)(a_long >> ((com.db4o.YapConst.LONG_BYTES
					 - 1 - i) * 8));
			}
		}

		internal static void writeLong(long a_long, byte[] bytes)
		{
			for (int i = 0; i < com.db4o.YapConst.LONG_BYTES; i++)
			{
				bytes[i] = (byte)(a_long >> ((com.db4o.YapConst.LONG_BYTES - 1 - i) * 8));
			}
		}

		internal static long readLong(com.db4o.YapWriter writer)
		{
			long l_return = 0;
			for (int i = 0; i < com.db4o.YapConst.LONG_BYTES; i++)
			{
				l_return = (l_return << 8) + (writer._buffer[writer._offset++] & unchecked((int)(
					0xff)));
			}
			return l_return;
		}

		protected long i_compareTo;

		internal virtual long val(object obj)
		{
			return ((long)obj);
		}

		internal override void prepareComparison1(object obj)
		{
			i_compareTo = val(obj);
		}

		internal override bool isEqual1(object obj)
		{
			return obj is long && val(obj) == i_compareTo;
		}

		internal override bool isGreater1(object obj)
		{
			return obj is long && val(obj) > i_compareTo;
		}

		internal override bool isSmaller1(object obj)
		{
			return obj is long && val(obj) < i_compareTo;
		}
	}
}
